//------------------------------------------------------------------------------
/*
    This file is part of Bottle2D.

    Bottle2D is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Bottle2D is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Bottle2D.  If not, see <http://www.gnu.org/licenses/>.
*/
//------------------------------------------------------------------------------
#ifndef GROUP_HPP
#define GROUP_HPP
//------------------------------------------------------------------------------
#include "Drawable.hpp"
#include "PhysicObject.hpp"
#include "Object.hpp"

#include "Event.hpp"
//------------------------------------------------------------------------------
#include <luabind/object.hpp>
//------------------------------------------------------------------------------
#include <vector>
//------------------------------------------------------------------------------
namespace En
{
//------------------------------------------------------------------------------
#define X(typeName, typeId) \
        class typeName;
    X_OBJ_LIST
#undef X
class ParticleModel;
//------------------------------------------------------------------------------
/*const bool g_drawGroupCenter = true;
const bool g_drawPhysic = true;*/
//------------------------------------------------------------------------------
bool compareZpos(const Drawable *d1, const Drawable *d2);
//------------------------------------------------------------------------------
/// LUA_HEADER_COPY_ON ///
/** A Group is a container wich can contain drawable and physic object.
  * When a group is moved or rotated, its child will be moved or rotated too. It's like a folder: if you Destroy it, its children will be destroyed too.
  * It could be created using Group::CreateGroup.
  */
/// LUA_HEADER_COPY_OFF ///
class Group : public Drawable
{
public:
    Group(Group *parent, const std::string &name);
    Group(const Group &grp, Group *newParent);
    virtual ~Group();

    static Group *Create(Group *parent, const std::string &name) {return new Group(parent, name);}

    /// LUA_EXPORT_ON(Group, Drawable) ///

    // Clone & Create
    /** Create a clone of the object and return an handler to the clone. */
    virtual Group *Clone() const {return CloneToGroup(NULL);}

    /** Create a clone of the object, change the parent and return an handler to the clone. */
    virtual Group *CloneToGroup(Group *newParent) const {Group *obj = new Group(*this, newParent); obj->RegisterClone(newParent); return obj;}

    // Scale
    /** Edit scale */
    //LUA// luaProperty(Vector2D, scale, GetScale, SetScale);

    /** Set the group scale. It allows you to modify the size of a single group and its childs. */
    void SetScale(const Vector2D &scale);

    /** Get the group scale. */
    Vector2D GetScale();

    /** Return the number of physic and drawable children that this group contains */
    int GetChildCount() const;

    /** Return the physic child number id */
    PhysicObject *GetPhysicChild(int id);

    /** Return the number of physic children that this group contains */
    int GetPhysicChildCount() const;

    /** Return the drawable child number id */
    Drawable *GetDrawableChild(int id);

    /** Return the number of drawable children that this group contains */
    int GetDrawableChildCount() const;

    /** Destroy every children of the group */
    void DestroyChildren();

    // Cut
    /** Enable / disable the cutting of a certain part of the group. Use Group::SetCutRect to define that part. */
    void EnableCut(bool isCut) {m_isCut = isCut;}

    /** Return true if the cutting is enabled, false otherwise. */
    bool IsCut() const {return m_isCut;}

    /** Set the cut rectangle. You should enable cutting. */
    void SetCutRect(const Vector2D &position, const Vector2D &size);

    /** Get the cut rectangle's position. */
    Vector2D GetCutRectPosition() const;

    /** Get the cut rectangle's size. */
    Vector2D GetCutRectSize() const;





    //void SetScale(float scale);

    // FindByName
    /*#define X(typeName, typeId) \
            typeName *FindByName ## typeName(const std::string &name);
        X_OBJ_LIST
    #undef X*/
    /** Get a lua table containing all its childs.\\n
      * Warning: no particular order are warranted.
      */
    luabind::object GetChildTable();

    /** Get a child from its index in the group (a number for 1 to Group::GetChildCount().\\n
      * Warning: no particular order are warranted.
      * Use Group::GetChildTable() to traverse a whole group : it will be faster.
      */
    luabind::object FindByIndex(int id);

    // FIXME: Accélérer la recherche avec des maps ?
    /** Find an object in the group with its name and cast it with the good type. Return nil if the object doesn't exist. */
    luabind::object FindByName(const std::string &name);

    /** Find an object in the group with its name. This function will also check the type of the object. */
    PhLine *FindByNamePhLine(const std::string &name);
    /** Find an object in the group with its name. This function will also check the type of the object. */
    PhRect *FindByNamePhRect(const std::string &name);
    /** Find an object in the group with its name. This function will also check the type of the object. */
    PhTriangle *FindByNamePhTriangle(const std::string &name);
    /** Find an object in the group with its name. This function will also check the type of the object. */
    PhCircle *FindByNamePhCircle(const std::string &name);
    /** Find an object in the group with its name. This function will also check the type of the object. */
    PhShape *FindByNamePhShape(const std::string &name);
    /** Find an object in the group with its name. This function will also check the type of the object. */
    Group *FindByNameGroup(const std::string &name);
    /** Find an object in the group with its name. This function will also check the type of the object. */
    ParticleSource *FindByNameParticleSource(const std::string &name);
    /** Find an object in the group with its name. This function will also check the type of the object. */
    Shape *FindByNameShape(const std::string &name);
    /** Find an object in the group with its name. This function will also check the type of the object. */
    Sprite *FindByNameSprite(const std::string &name);
    /** Find an object in the group with its name. This function will also check the type of the object. */
    Text *FindByNameText(const std::string &name);
    /** Find an object in the group with its name. This function will also check the type of the object. */
    PostFX *FindByNamePostFX(const std::string &name);
    /** Find an object in the group with its name. This function will also check the type of the object. */
    Screen *FindByNameScreen(const std::string &name);
    /** Find an object in the group with its name. This function will also check the type of the object. */
    ScreenRenderer *FindByNameScreenRenderer(const std::string &name);

    // Create
    /** Create a new physic line and attach it to this group.\\n
      * \\param phGroupId The physic group of the object, in order to generate colision events (see Screen::AddPhysicGroup()).
      * \\param name The name of the object.
      * \\param pt1 First point of the line.
      * \\param pt2 Second point of the line.
      * \\return The created object.
      */
    PhLine *CreatePhLine(int phGroupId, const std::string &name, const Vector2D &pt1, const Vector2D &pt2);

    /** Create a new physic rectangle and attach it to this group.\\n
      * \\param phGroupId The physic group of the object, in order to generate colision events (see Screen::AddPhysicGroup()).
      * \\param name The name of the object.
      * \\param size The size of the rectangle. Use Object::SetPosition() to change the object's position.
      * \\return The created object.
      */
    PhRect *CreatePhRect(int phGroupId, const std::string &name, const Vector2D &size);

    /** Create a new physic triangle and attach it to this group.\\n
      * \\param phGroupId The physic group of the object, in order to generate colision events (see Screen::AddPhysicGroup()).
      * \\param name The name of the object.
      * \\param pt1 First point of the triangle.
      * \\param pt2 Second point of the triangle.
      * \\param pt3 Third point of the triangle.
      * \\return The created object.
      */
    PhTriangle *CreatePhTriangle(int phGroupId, const std::string &name, const Vector2D &pt1, const Vector2D &pt2, const Vector2D &pt3);

    /** Create a physic shape and attach it to this group.\\n
      * When a shape is created, you can add points to it by calling PhShape::AddPoint(). The shape must be convex !
      * \\param phGroupId The physic group of the object, in order to generate colision events (see Screen::AddPhysicGroup()).\\n
      * \\param name The name of the object.
      * \\return The created object.
      */
    PhShape *CreatePhShape(int phGroupId, const std::string &name);

    /** Create a new physic disk and attach it to this group.\\n
      * \\param phGroupId The physic group of the object, in order to generate colision events (see Screen::AddPhysicGroup()).
      * \\param name The name of the object.
      * \\param center The center of the circle. Can be changed with Object::SetPosition().
      * \\param radius The radius of the disk.
      * \\return The created object.
      */
    PhCircle *CreatePhCircle(int phGroupId, const std::string &name, const Vector2D &center, float radius);

    /** Create a Group object and add it to the current group.\\n
      * \\param name The name of the object.
      * \\return The created object.
      */
    Group *CreateGroup(const std::string &name);

    /** Create a ParticleSource object and add it to the current group.\\n
      * \\param name The name of the object.
      * \\param model The particle model for this ParticleSource.
      * \\param maxParticleNumber Used to reserve memory. It should not be oversized (unusefull memory would be allocated) or undersized (some particle won't be created)
      * <ul>
      *     <li>If your particle source just has a flow and an unlimited tank, then \\a maxParticleNumber should be \\a flow * \\a maxLifeTime.</li>
      *     <li>If your particle is an explosion, and all particle are dead before starting a new explosion, then \\a maxParticleNumber should be \\a tank.</li>
      * </ul>
      * \\return The created object.
      */
    ParticleSource *CreateParticleSource(const std::string &name, const ParticleModel &model, int maxParticleNumber);

    /** Create a Shape object and add it to the current group.\\n
      * \\param name The name of the object.
      * \\return The created object.
      */
    Shape *CreateShape(const std::string &name);

    /** Create a shape, initialise it as a line and add it to the current group.\\n
      * \\param name The name of the object.
      * \\param P1 Position of the first point
      * \\param P2 Position second point
      * \\param Thickness Line thickness
      * \\param Col Color used to draw the line
      * \\param Outline Outline width (0 by default)
      * \\param OutlineCol Color used to draw the outline
      * \\return The created object.
      */
    Shape *CreateShapeLine(const std::string &name, const Vector2D& P1, const Vector2D& P2, float Thickness, const Color& Col, float Outline, const Color& OutlineCol);

    /** Create a shape, initialise it as a rectangle and add it to the current group.\\n
      * \\param name The name of the object.
      * \\param P1 Position of the first point
      * \\param P2 Position second point
      * \\param Col Color used to fill the rectangle
      * \\param Outline Outline width
      * \\param OutlineCol Color used to draw the outline
      * \\return The created object.
      */
    Shape *CreateShapeRectangle(const std::string &name, const Vector2D& P1, const Vector2D& P2, const Color& Col, float Outline, const Color& OutlineCol);

    /** Create a circle / disk, initialise it as a rectangle and add it to the current group.\\n
      * \\param name The name of the object.
      * \\param Center Position of the center
      * \\param Radius Radius
      * \\param Col Color used to fill the circle
      * \\param Outline Outline width
      * \\param OutlineCol Color used to draw the outline
      * \\return The created object.
      */
    Shape *CreateShapeCircle(const std::string &name, const Vector2D& Center, float Radius, const Color& Col, float Outline, const Color& OutlineCol);

    /** Create a Sprite object and add it to the current group.\\n
      * \\param name The name of the object.
      * \\return The created object.
      */
    Sprite *CreateSprite(const std::string &name);

    /** Create a Text object and add it to the current group.\\n
      * \\param name The name of the object.
      * \\param text Text of the object.
      * \\return The created object.
      */
    Text *CreateText(const std::string &name, const std::string &text);

    /** Create a PostFX object and add it to the current group.\\n
      * \\param name The name of the object.
      * \\return The created object.
      */
    PostFX *CreatePostFX(const std::string &name);

    /** Create a ScreenRenderer object and add it to the current group.\\n
      * \\param name The name of the object.
      * \\return The created object.
      */
    ScreenRenderer *CreateScreenRenderer(const std::string &name);

    // Event auto remove
    void RegisterEventAutoRemoveOnDestroy(EventHandlerBase *event);
    /** Register an event to the group so that it will be destroyed when the group is destroyed.\\n
      * This can be useful if some event can't work without an object and can create segmentation fault when they try to use a removed object.\\n
      * This must be used with caution : if the event is already removed when the group is Destroyed, a segmentation fault will occure.\\n
      * Do not regiser one event on more than one group !
     */

    /// LUA_EXPORT_OFF ///

    // Not to use with Lua
    // Lua
    //static void RegisterClassLua(lua_State *myLuaState);

    // Rotation
    void PropagateRotation(float angle);

    // Add / Remove
    int Add(Drawable *obj);
    void Add(PhysicObject *obj);

    void Remove(Drawable *obj);
    void Remove(PhysicObject *obj);

    // Clean Deleted
    virtual void CleanDeleted();
    void AddToDelete(Object *obj);

    // Register clone
    virtual void RegisterClone(Group *newParent);

    // Type
    ADD_GET_TYPE_FN(OBJ_TYPE_DRAWABLE_GROUP)

    // Zpos
    //void RefreshZpos();

    virtual void SetRotation(float angle);
    virtual void ProceedMvt(float time);
    virtual void Render(sf::RenderTarget &App);

    // Some object need to be noticed when the parent screen is changed
    virtual void OnChangeScreen(Screen *backScreen, Screen *newScreen);
protected:
    Object *FindByNameUncast(const std::string &name, const size_t &id = 0);
    void Clear();

    // Cast an unknown object for a lua object
    luabind::object CastObject(Object *obj);

protected:
    //std::vector<Drawable*> m_drawTbl;
    typedef std::map<int, Drawable* > DrawLowZPosMap;
    typedef std::map<int, DrawLowZPosMap > DrawZPosMap;

    DrawZPosMap m_drawTbl;
    std::vector<PhysicObject*> m_phTbl;

    std::vector<Object*> m_toDelete;

    Vector2D m_scale;

    // Cut
    sf::FloatRect m_cutRect;
    bool m_isCut;

    // RegisterEventAutoRemoveOnDestroy
    std::vector<EventHandlerBase*> m_eventToBeDestroyedOnDelete;
};
//------------------------------------------------------------------------------
} // namespace En
//------------------------------------------------------------------------------
#endif // GROUP_HPP
