//------------------------------------------------------------------------------
/*
    This file is part of Bottle2D.

    Bottle2D is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Bottle2D is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Bottle2D.  If not, see <http://www.gnu.org/licenses/>.
*/
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
#include "Screen.hpp"
#include "Global.hpp"
#include <iostream>
//------------------------------------------------------------------------------
#include "MyException.hpp"
#include "nbStrConvert.hpp"
#include "Event.hpp"
#include "PseudoEvent.hpp"
//------------------------------------------------------------------------------
namespace En
{
//------------------------------------------------------------------------------
using namespace std;
//------------------------------------------------------------------------------
/** Create a basic config with a 1024*768 size, and set according ratio */
ScreenConfig::ScreenConfig() : isChanged(true), cameraBottomRight(Vector2D(1024, 768)), cameraTopLeft(Vector2D(0, 0)), screenSize(Vector2D(1024, 768)), isFixedRatio(true)
{
    forcedRatio = 1024.f / 768.f;
}
//------------------------------------------------------------------------------
/** Create a basic config with user size, and set according ratio */
ScreenConfig::ScreenConfig(const Vector2D &size) : isChanged(true), cameraBottomRight(size), cameraTopLeft(Vector2D(0, 0)), screenSize(size), isFixedRatio(true)
{
    forcedRatio = size.x / size.y;
}
//------------------------------------------------------------------------------
/** Set the screen size and the according ratio. Reset the camera */
void ScreenConfig::ChangeSize(const Vector2D &size)
{
    forcedRatio = size.x / size.y;
    screenSize = size;
    cameraTopLeft = Vector2D(0, 0);
    cameraBottomRight = size;

    isChanged = true;
}
//------------------------------------------------------------------------------
/** Reset the camera to show the entire screen */
void ScreenConfig::CameraReset()
{
    cameraTopLeft = Vector2D(0, 0);
    cameraBottomRight = screenSize;

    isChanged = true;
}
//------------------------------------------------------------------------------
/** Move the camera with the given offset */
void ScreenConfig::CameraMove(const Vector2D &mvt)
{
    cameraTopLeft += mvt;
    cameraBottomRight += mvt;

    isChanged = true;
}
//------------------------------------------------------------------------------

/** Zoom */
void ScreenConfig::CameraZoom(float addedZoom)
{
    CameraSetHalfSize(CameraGetHalfSize() / addedZoom);

    isChanged = true;
}
//------------------------------------------------------------------------------

/** Set zoom, using screen size */
void ScreenConfig::CameraSetZoom(float zoom)
{
    CameraReset();
    CameraZoom(zoom);

    isChanged = true;
}
//------------------------------------------------------------------------------
void ScreenConfig::CameraZoom(Vector2D addedZoom)
{
    Vector2D halfSize = CameraGetHalfSize();
    CameraSetHalfSize(Vector2D(halfSize.x / addedZoom.x, halfSize.y / addedZoom.y));

    isChanged = true;
}
//------------------------------------------------------------------------------
void ScreenConfig::CameraSetZoom(Vector2D zoom)
{
    CameraReset();
    CameraZoom(zoom);

    isChanged = true;
}
//------------------------------------------------------------------------------
/** Get zoom, using screen size */
/*float ScreenConfig::CameraGetZoom()
{

}*/
//------------------------------------------------------------------------------
/** Set the camera center */
void ScreenConfig::CameraSetCenter(const Vector2D &center)
{
    CameraMove(CameraGetCenter() - center);

    isChanged = true;
}
//------------------------------------------------------------------------------

/** Set the camera half size */
void ScreenConfig::CameraSetHalfSize(const Vector2D &halfSize)
{
    Vector2D center = CameraGetCenter();
    cameraTopLeft = center - halfSize;
    cameraBottomRight = center + halfSize;

    isChanged = true;
}
//------------------------------------------------------------------------------
/** Get the camera center */
Vector2D ScreenConfig::CameraGetCenter()
{
    return Vector2D((cameraBottomRight.x + cameraTopLeft.x) / 2.f, (cameraBottomRight.y + cameraTopLeft.y) / 2.f);
}
//------------------------------------------------------------------------------

/** Get the camera half size */
Vector2D ScreenConfig::CameraGetHalfSize()
{
    return Vector2D((cameraBottomRight.x - cameraTopLeft.x) / 2.f, (cameraBottomRight.y - cameraTopLeft.y) / 2.f);
}
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
Screen::Screen(const std::string &name) : Group(NULL, name), m_isMouseRelative(false)/*, m_showFps(false)*/, m_timeSpeed(1), m_backgroundColor(Color(0, 0, 0)), m_savedWindowSize(sf::Vector2i(-1, -1)), m_isCursorVisible(true)
{
    Global::AddScreen(this);
    RefreshStripView(Global::GetApp());
}
//------------------------------------------------------------------------------
EventHandlerBase* Screen::RegisterButtonEvent(const luabind::object &luaFunction, const std::string &keyName, bool pressed)
{
    EventHandler *evt = new ButtonEventHandler(luaFunction, keyName, pressed);
    m_evtHandlerTbl.push_back(evt);
    return evt;
}
//------------------------------------------------------------------------------

EventHandlerBase* Screen::RegisterMouseMoveEvent(const luabind::object &luaFunction)
{
    EventHandler *evt = new MouseMoveEventHandler(luaFunction);
    m_evtHandlerTbl.push_back(evt);
    return evt;
}
//------------------------------------------------------------------------------
EventHandlerBase *Screen::RegisterButtonInZoneEvent(const luabind::object &luaFunction, MouseZoneEvent *zoneEvt, const std::string &keyName, bool pressed)
{
    ButtonInZoneEventHandler *evt = new ButtonInZoneEventHandler(luaFunction, zoneEvt, keyName, pressed);
    m_evtHandlerTbl.push_back(evt);
    return evt;
}
//------------------------------------------------------------------------------
Timer *Screen::RegisterTimer(const luabind::object &luaFunction, float interval, bool isRunning)
{
    Timer *evt = new Timer(luaFunction, interval, isRunning);
    m_pseudoEvtHandlerTbl.push_back(evt);
    return evt;
}
//------------------------------------------------------------------------------
Chronometer *Screen::RegisterChronometer(const luabind::object &luaFunction, float totalTime, bool isRunning)
{
    Chronometer *evt = new Chronometer(luaFunction, totalTime, isRunning);
    m_pseudoEvtHandlerTbl.push_back(evt);
    return evt;
}
//------------------------------------------------------------------------------
MouseZoneEvent *Screen::RegisterMouseZoneEvent(const luabind::object &luaFunction, PhysicObject *zone)
{
    MouseZoneEvent *evt = new MouseZoneEvent(luaFunction, zone);
    m_pseudoEvtHandlerTbl.push_back(evt);
    return evt;
}
//------------------------------------------------------------------------------
EventHandlerBase *Screen::RegisterFrameEvent(const luabind::object &luaFunction)
{
    FrameEvent *evt = new FrameEvent(luaFunction);
    m_pseudoEvtHandlerTbl.push_back(evt);
    return evt;
}
//------------------------------------------------------------------------------
void Screen::DrawRatioStrip(sf::RenderTarget &App)
{
    App.Draw(m_strip1);
    App.Draw(m_strip2);
}
//------------------------------------------------------------------------------
void Screen::RefreshStripView(sf::RenderTarget &App)
{
    if(!m_screenCfg.IsFixedRatio())
    {
        return;
    }

    // Hauteur ralative à la taille totale à enlever pour atteindre le goal ratio :
    float hDiffRatio = (float(App.GetHeight()) - float(App.GetWidth()) / m_screenCfg.GetForcedRatio()) / float(App.GetHeight());

    if(hDiffRatio >= 0) // il faut ajouter des bandes horizontales
    {
        //sf::FloatRect rect(0, 0, m_screenCfg.screenSize.x, m_screenCfg.screenSize.y / (1.f - hDiffRatio));
        sf::FloatRect rect(m_screenCfg.GetCameraTopLeft().x, m_screenCfg.GetCameraTopLeft().y, m_screenCfg.GetCameraBottomRight().x, m_screenCfg.GetCameraBottomRight().y / (1.f - hDiffRatio));
        float width = rect.GetWidth();
        float height = rect.GetHeight();

        //m_isVerticalStrip = false;
        float stripHeight = hDiffRatio * height / 2.f;
        m_strip1 = sf::Shape::Rectangle(Vector2D(0, 0), Vector2D(width, stripHeight), GetBackgroundColor(), 0);
        m_strip1.Move(m_screenCfg.GetCameraTopLeft());
        m_strip2 = sf::Shape::Rectangle(Vector2D(0, height - stripHeight), Vector2D(width, height), GetBackgroundColor(), 0);
        m_strip2.Move(m_screenCfg.GetCameraTopLeft());

        // Screen position
        SetPosition(Vector2D(0, stripHeight));

        // View
        m_screenView.SetFromRect(rect);
    }
    else // Bandes verticales
    {
        float vDiffRatio = (float(App.GetWidth()) - float(App.GetHeight()) * m_screenCfg.GetForcedRatio()) / float(App.GetWidth());

        sf::FloatRect rect(m_screenCfg.GetCameraTopLeft().x, m_screenCfg.GetCameraTopLeft().y, m_screenCfg.GetCameraBottomRight().x / (1.f - vDiffRatio), m_screenCfg.GetCameraBottomRight().y);
        float width = rect.GetWidth();
        float height = rect.GetHeight();

        //m_isVerticalStrip = true;
        float stripWidth = vDiffRatio * width / 2.f;
        m_strip1 = sf::Shape::Rectangle(Vector2D(0, 0), Vector2D(stripWidth, height), GetBackgroundColor(), 0);
        m_strip1.Move(m_screenCfg.GetCameraTopLeft());
        m_strip2 = sf::Shape::Rectangle(Vector2D(width - stripWidth, 0), Vector2D(width, height), GetBackgroundColor(), 0);
        m_strip2.Move(m_screenCfg.GetCameraTopLeft());

        // Screen position
        SetPosition(Vector2D(stripWidth, 0));

        // View
        m_screenView.SetFromRect(rect);
    }

    m_screenCfg.isChanged = false;
    m_savedWindowSize = sf::Vector2i(App.GetWidth(), App.GetHeight());
}

//------------------------------------------------------------------------------
/*void Screen::RemoveEvent(EventHandler* event)
{
    std::vector<EventHandler*>::iterator it = find(m_evtHandlerTbl.begin(), m_evtHandlerTbl.end(), event);
    if(it == m_evtHandlerTbl.end())
    {
        MY_THROW("RemoveEvent: not found !");
    }

    cout << it - m_evtHandlerTbl.begin() << endl;

    delete (*it);
    m_evtHandlerTbl.erase(it);
}
//------------------------------------------------------------------------------
void Screen::RemoveEvent(PseudoEventHandler* event)
{
    std::vector<PseudoEventHandler*>::iterator it = find(m_pseudoEvtHandlerTbl.begin(), m_pseudoEvtHandlerTbl.end(), event);
    if(it == m_pseudoEvtHandlerTbl.end())
    {
        MY_THROW("RemoveEvent: not found !");
    }

    m_pseudoEvtHandlerTbl.erase(it);
}*/
//------------------------------------------------------------------------------
void Screen::AddPhysicGroup(int phGroupId1, int phGroupId2)
{
    m_world.AddPhysicGroup(phGroupId1, phGroupId2);
}
//------------------------------------------------------------------------------
Color Screen::GetBackgroundColor()
{
    return m_backgroundColor;
}
//------------------------------------------------------------------------------
void Screen::SetBackgroundColor(const Color &color)
{
    m_backgroundColor = color;
    m_savedWindowSize.x = -1; // Refresh shape
}
//------------------------------------------------------------------------------
/** Get the screen configuration */
ScreenConfig &Screen::GetScreenConfig()
{
    return m_screenCfg;
}
//------------------------------------------------------------------------------
/** Set the screen configuration */
void Screen::SetScreenConfig(const ScreenConfig &screenConfig)
{
    m_screenCfg = screenConfig;
}

//------------------------------------------------------------------------------

void Screen::PlaySound(Sound &sound)
{
    std::list<sf::Sound>::iterator it;
    for(it = m_sndTbl.begin(); it != m_sndTbl.end(); it++ )
    {
        if(it->GetStatus() != sf::Sound::Playing)
        {
            it->SetBuffer(sound);
            it->Play();
            break;
        }
    }

    if(it == m_sndTbl.end()) // Pas de libre trouvé
    {
        m_sndTbl.push_back(sf::Sound());
        m_sndTbl.back().SetBuffer(sound);
        m_sndTbl.back().Play();
    }
    else // Clean
    {
        for(it = m_sndTbl.begin(); it != m_sndTbl.end(); it++ )
        {
            if(it->GetStatus() != sf::Sound::Playing)
            {
                it = m_sndTbl.erase(it);
                it--; // On decrease de 1 comme on l'ajoute dans la bouble for.
            }
        }
    }
}
//------------------------------------------------------------------------------
void Screen::SetTimeSpeed(float timeSpeed)
{
    m_timeSpeed = timeSpeed;
}
//------------------------------------------------------------------------------
float Screen::GetTimeSpeed() const
{
    return m_timeSpeed;
}
//------------------------------------------------------------------------------
void Screen::Run()
{
    sf::RenderWindow &App = Global::GetApp();

    // Physic
    ProceedMvt(App.GetFrameTime() * m_timeSpeed);
    GetPhysicWorld()->ProceedIntersection();

    // Event
    ProceedEvent();

    // Clean des éléments supprimés
    CleanDeleted();
}
//------------------------------------------------------------------------------
void Screen::Render(sf::RenderTarget &App)
{
    // Strips & view
    if(m_screenCfg.isChanged || m_savedWindowSize != sf::Vector2i(App.GetWidth(), App.GetHeight()))
    {
        RefreshStripView(App);
    }

    App.SetView(m_screenView);

    Group::Render(App);
    DrawRatioStrip(App);
}
//------------------------------------------------------------------------------
void Screen::ProceedEvent()
{
    // Pseudo event

    // FIXME : Liste à déléter à faire ! // Ou pas
    for(std::vector<PseudoEventHandler*>::iterator it = m_pseudoEvtHandlerTbl.begin(); it != m_pseudoEvtHandlerTbl.end(); it++)
    {
        if((*it)->ShouldBeDestroyed())
        {
            delete *it;
            it = m_pseudoEvtHandlerTbl.erase(it) - 1;
        }
        else
        {
            // We need to keep the pos in case event are created in the event function (iterators would be invalidated)
            size_t pos = it - m_pseudoEvtHandlerTbl.begin();
            (*it)->Match(this, Global::GetApp());
            it = m_pseudoEvtHandlerTbl.begin() + pos;
        }
    }

    // Event
    sf::Event event;
    while(Global::GetApp().GetEvent(event))
    {
        // FIXME : Liste à déléter à faire !
        for(std::vector<EventHandler*>::iterator it = m_evtHandlerTbl.begin(); it != m_evtHandlerTbl.end(); it++)
        {
            if((*it)->ShouldBeDestroyed())
            {
                delete *it;
                it = m_evtHandlerTbl.erase(it) - 1;
            }
            else
            {
                // We need to keep the pos in case event are created in the event function (iterators would be invalidated)
                size_t pos = it - m_evtHandlerTbl.begin();
                (*it)->Match(event, this, Global::GetApp());
                it = m_evtHandlerTbl.begin() + pos;
            }
        }

        // FIXME : cout à enlever
        if(IsMouseRelative())
        {
            if(event.Type == sf::Event::GainedFocus)
            {
                SetMouseRelative(true, false);
            }
            else if(event.Type == sf::Event::LostFocus)
            {
                SetMouseRelative(false, false);
            }
        }

        if(event.Type == sf::Event::Closed)
        {
            Global::Quit();
        }
    }

    if(IsMouseRelative())
    {
        Global::GetApp().SetCursorPosition(Global::GetApp().GetWidth() / 2, Global::GetApp().GetHeight() / 2);
    }
}
//------------------------------------------------------------------------------
Screen::~Screen()
{
    // !! We need to destroy groups before events : some of them might try to destroy event when they are destroyed.
    // Disconect every object
    OnChangeScreen(this, NULL);

    // On doit supprimer le contenu du groupe avant que Screen ne devienne un simple groupe
    Clear();

    // Delete Event Handler
    for(std::vector<EventHandler*>::iterator it = m_evtHandlerTbl.begin(); it != m_evtHandlerTbl.end(); it++ )
    {
        delete *it;
    }
    m_evtHandlerTbl.clear();

    for(std::vector<PseudoEventHandler*>::iterator it = m_pseudoEvtHandlerTbl.begin(); it != m_pseudoEvtHandlerTbl.end(); it++ )
    {
        delete *it;
    }
    m_pseudoEvtHandlerTbl.clear();
}
//------------------------------------------------------------------------------
void Screen::OnGetFocus()
{
    // Sound
    for(std::list<sf::Sound>::iterator it = m_sndTbl.begin(); it != m_sndTbl.end(); it++ )
    {
        if(it->GetStatus() == sf::Sound::Paused) // Pas les stoped
        {
            it->Play();
        }
    }

    // Mouse
    if(IsMouseRelative())
    {
        SetMouseRelative(true, false);
    }
    ShowCursor(m_isCursorVisible);
}
//------------------------------------------------------------------------------
void Screen::OnLoseFocus()
{
    // Sound
    for(std::list<sf::Sound>::iterator it = m_sndTbl.begin(); it != m_sndTbl.end(); it++ )
    {
        if(it->GetStatus() == sf::Sound::Playing)
        {
            it->Pause();
        }
    }

    // Mouse
    if(IsMouseRelative())
    {
        SetMouseRelative(false, false);
    }

    // To do not have a deleted View
    sf::RenderWindow &App = Global::GetApp();
    App.SetView(App.GetDefaultView());
}
//------------------------------------------------------------------------------
void Screen::ChangeParent(Group *parent)
{
    MY_THROW("ChangeParent for Screen");
}
//------------------------------------------------------------------------------
void Screen::SetMouseRelative(bool relative, bool changeValue)
{
    if(!m_isMouseRelative && relative) // Change absolute to relative
    {
        m_backMousePos = sf::Vector2<int>(Global::GetApp().GetInput().GetMouseX(), Global::GetApp().GetInput().GetMouseY());
        Global::GetApp().SetCursorPosition(Global::GetApp().GetWidth() / 2, Global::GetApp().GetHeight() / 2); // Pr Eviter de heurter les bords.
    }
    else if(m_isMouseRelative && !relative) // Change relative to absolute
    {
        Global::GetApp().SetCursorPosition(m_backMousePos.x, m_backMousePos.y);
    }

    if(changeValue)
    {
        m_isMouseRelative = relative;
    }
}
//------------------------------------------------------------------------------
bool Screen::IsMouseRelative() const
{
    return m_isMouseRelative;
}
//------------------------------------------------------------------------------
void Screen::ShowCursor(bool show)
{
    m_isCursorVisible = show;
    Global::GetApp().ShowMouseCursor(m_isCursorVisible);
}
//------------------------------------------------------------------------------
bool Screen::IsCursorVisible() const
{
    return m_isCursorVisible;
}
//------------------------------------------------------------------------------
/** Get the number of event and pseudo event registered in this screen. */
int Screen::GetEventCount() const
{
    return m_evtHandlerTbl.size() + m_pseudoEvtHandlerTbl.size();
}
//------------------------------------------------------------------------------
} // namespace En
//------------------------------------------------------------------------------
