//------------------------------------------------------------------------------
/*
    This file is part of Bottle2D.

    Bottle2D is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Bottle2D is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Bottle2D.  If not, see <http://www.gnu.org/licenses/>.
*/
//------------------------------------------------------------------------------
#ifndef MAINGROUP_HPP
#define MAINGROUP_HPP

#include "PhysicWorld.hpp"
#include "Group.hpp"
#include "Global.hpp"
#include <SFML/Audio.hpp>
//------------------------------------------------------------------------------
#include "PseudoEvent.hpp"
#include "Event.hpp"
//------------------------------------------------------------------------------
#include <iostream>
#include <list>
//------------------------------------------------------------------------------
namespace En
{
//------------------------------------------------------------------------------
/// LUA_HEADER_COPY_ON ///
/** A ScreenConfig is associated to a screen. It alows you to chose the screen virtual size and its ratio. */
/// LUA_HEADER_COPY_OFF ///
//------------------------------------------------------------------------------
class ScreenConfig
{
public:
    /// LUA_EXPORT_ON(ScreenConfig) ///
    // Constructor
    /** Create a basic config with a 1024*768 size, and set according ratio */
    ScreenConfig();

    /** Create a basic config with user size, and set according ratio */
    ScreenConfig(const Vector2D &size);

    /** Set the screen size and the according ratio. Reset the camera */
    void ChangeSize(const Vector2D &size);

    /** Reset the camera to show the entire screen */
    void CameraReset();

    // Lua
    /** Access the ratio */
    //LUA// luaProperty(float, forcedRatio, GetForcedRatio, SetForcedRatio);
    /** Set the ratio */
    void SetForcedRatio(float value) {forcedRatio = value; isFixedRatio = true; isChanged = true;}
    /** Get the ratio */
    float GetForcedRatio() const {return forcedRatio;}

    /** Choose if the engine should always use the same ratio on resize (add strips to the screen in order to keep the same ratio) */
    //LUA// luaProperty(bool, isFixedRatio, IsFixedRatio, SetFixedRatio);
    /** Choose if the engine should always use the same ratio on resize (add strips to the screen in order to keep the same ratio) */
    void SetFixedRatio(bool value) {isFixedRatio = value; isChanged = true;}
    /** Tells if the engine should always use the same ratio on resize */
    bool IsFixedRatio() const {return isFixedRatio;}

    /** Set / Get the screen size but doesn't change the ratio */
    //LUA// luaProperty(Vector2D, screenSize, GetScreenSize, SetScreenSize);

    /** Set the screen size but doesn't change the ratio */
    void SetScreenSize(const Vector2D &value) {screenSize = value; isChanged = true;}

    /** Get the screen size */
    Vector2D GetScreenSize() const {return screenSize;}

    // Camera
    /** Set the camera top left position */
    //LUA// luaProperty(Vector2D, cameraTopLeft, GetCameraTopLeft, SetCameraTopLeft);
    /** Set the camera top left position */
    void SetCameraTopLeft(const Vector2D &value) {cameraTopLeft = value; isChanged = true;}
    /** Get the camera top left position */
    Vector2D GetCameraTopLeft() const {return cameraTopLeft;}


    /** Set the camera bottom right position */
    //LUA// luaProperty(Vector2D, cameraBottomRight, GetCameraBottomRight, SetCameraBottomRight);
    /** Set the camera bottom right position */
    void SetCameraBottomRight(const Vector2D &value) {cameraBottomRight = value; isChanged = true;}
    /** Get the camera bottom right position */
    Vector2D GetCameraBottomRight() const {return cameraBottomRight;}

    /** Move the camera with the given offset */
    void CameraMove(const Vector2D &mvt);

    /** Zoom */
    void CameraZoom(float addedZoom);

    /** Zoom */
    void CameraZoom(Vector2D addedZoom);

    // /** Access the camera zoom */
    // //LUA// luaProperty(float, zoom, CameraGetZoom, CameraSetZoom);

    /** Set zoom, using screen size as a reference */
    void CameraSetZoom(float zoom);

    /** Set zoom, using screen size as a reference */
    void CameraSetZoom(Vector2D zoom);

    // /** Get zoom, using screen size */
    // float CameraGetZoom();

    /** Access the camera half size */
    //LUA// luaProperty(Vector2D, cameraCenter, CameraGetCenter, CameraSetCenter);

    /** Set the camera center */
    void CameraSetCenter(const Vector2D &center);

    /** Get the camera center */
    Vector2D CameraGetCenter();

    /** Access the camera half size */
    //LUA// luaProperty(Vector2D, cameraHalfSize, CameraGetHalfSize, CameraSetHalfSize);

    /** Set the camera half size */
    void CameraSetHalfSize(const Vector2D &halfSize);

    /** Get the camera half size */
    Vector2D CameraGetHalfSize();
    /// LUA_EXPORT_OFF ///

    // NOT accessible in lua
    bool isChanged;
protected:
    Vector2D cameraBottomRight;
    Vector2D cameraTopLeft;
    Vector2D screenSize;
    bool isFixedRatio;
    float forcedRatio;
};
//------------------------------------------------------------------------------
/// LUA_HEADER_COPY_ON ///
/** A screen represent a part of your game where differenent physic and drawable object will evolve. You can also attach event to your screen.\\n
  * A bottle 2D app will always need a screen to run, but it is recomanded to use more than one. For example, you can create a menu screen and a game screen.\\n
  * Then, when you can switch to the game / menu by calling ChangeScreen(). When a screen is not running, it is freezed : no sound, physic and events are proceeded.\\n
  * Screens are created using the global function CreateScreen().
  */
/// LUA_HEADER_COPY_OFF ///
class Screen : public Group
{
public:
    Screen(const std::string &name = "mainGroup");
    static Screen *Create(const std::string &name) {return new Screen(name);}
    virtual ~Screen();

    /*void RemoveEvent(EventHandlerBase *event);
    void RemoveEvent(PseudoEventHandlerBase *event);*/

    /// LUA_EXPORT_ON(Screen, Group) ///
    // Event

    /** Create a new button event.
      * The lua function luaFunction will be called each time the button keyName is pressed (if pressed = true) or released (if pressed = false).\\n
      * Prototype of the called lua function : function luaFunction(keyName : string, pressedStatus : bool, thisEvent : EventHandlerBase).\\n
      * Note: if keyName is empty (keyName = ""), then the lua function will be called each time a button is pressed / released.\\n
      * Note2: The buttons could be mouse buttons, keyboard buttons and even joystick buttons.\\n
      * See \\ref page_keyList for the list of keys.
      */
    EventHandlerBase *RegisterButtonEvent(const luabind::object &luaFunction, const std::string &keyName, bool pressed);

    /** Create a new button in zone event.
      * The lua function luaFunction will be called each time the button keyName is pressed (if pressed = true) or released (if pressed = false) and the mouse is in the zone corresponding to the zoneEvt.\\n
      * Parameters of the called lua function : the key name (string), the pressed status (bool), the MouseZoneEvent which corresponds to the zone and the EventHandlerBase.\\n
      * Note: if keyName is empty (keyName = ""), then the lua function will be called each time a button is pressed / released.\\n
      * Note2: The buttons could be mouse buttons, keyboard buttons and even joystick buttons.\\n
      * See \\ref page_keyList for the list of keys.
      */
    EventHandlerBase *RegisterButtonInZoneEvent(const luabind::object &luaFunction, MouseZoneEvent *zoneEvt, const std::string &keyName, bool pressed);

    /** Create a new mouse move event.
      * The lua function luaFunction will be called each time the mouse is moved.\\n
      * Parameters of the called lua function : the mouse position (Vector2D) and this EventHandlerBase.\\n
      * Note: if the mouse is configured as relative (cf: SetMouseRelative), then the position argument is the relative mouse mouvement.
      */
    EventHandlerBase *RegisterMouseMoveEvent(const luabind::object &luaFunction);


    // Pseudo event
    /**
      * Create a new Timer.
      * The lua function luaFunction will be called at each interval while the timer is still running.\\n
      * Parameters of the called lua function : this Timer.
      */
    Timer *RegisterTimer(const luabind::object &luaFunction, float interval, bool isRunning = true);

    /** Create a new chronometer.
      * The lua function luaFunction will be called at each frame during totalTime. It is  very useful to create mouvement.\\n
      * Parameters of the called lua function : the elapsed time ratio (float) and this Chronometer.\\n
      * Note: The engine ensure that the function receive a ratio of 1 at the end of the chronometer.
      */
    Chronometer *RegisterChronometer(const luabind::object &luaFunction, float totalTime, bool isRunning = true);

    /** Create a new mouse in zone event.
      * The lua function luaFunction will be called each time the mouse enter or leave the PhysicObject zone.\\n
      * Parameters of the called lua function : this MouseZoneEvent.\\n
      * Note : Use MouseZoneEvent::IsInside() to see if the mouse is currently inside
      */
    MouseZoneEvent *RegisterMouseZoneEvent(const luabind::object &luaFunction, PhysicObject *zone);

    /** Create a new Frame Event.
      * The lua function luaFunction will be called at each frame with the frame time (time between the last frame) and this eventHandler as a parameter
      */
    EventHandlerBase *RegisterFrameEvent(const luabind::object &luaFunction);

    // Sound
    /** Play a the sound. A sound is linked to a screen. If you change screen, the sound will be paused and will resume when you select back the screen. */
    void PlaySound(Sound &sound);

    // Mouse
    /** Turn on / off the mouse relative mode (cf: Screen::RegisterMouseMoveEvent) */
    void SetMouseRelative(bool relative) {SetMouseRelative(relative, true);}

    /** Tels if the mouse is in a relative or absolute mode */
    bool IsMouseRelative() const;

    /** Show / hide the mouse cursor */
    void ShowCursor(bool show);

    /** Tels if the mouse cursor is shown */
    bool IsCursorVisible() const;

    /** Get the number of event and pseudo event registered in this screen. */
    int GetEventCount() const;


    // FPS
    // /** Show the FPS (for optimisation) */
    //void ShowFps(bool show = false) {m_showFps = show;}

    // /** Tels if FPS ate shown */
    // bool IsFpsVisible() {return m_showFps;}

    // TimeSpeed
    /** Access current time speed. Elapsed time = RealTime * timeSpeed */
    //LUA// luaProperty(float, timeSpeed, GetTimeSpeed, SetTimeSpeed);

    /** Set time speed */
    void SetTimeSpeed(float timeSpeed);

    /** Get time speed */
    float GetTimeSpeed() const;

    // PhysicWorld
    /** Add a physic grpoup.
      * To use physic intersecion events, you have to add some "physic group".\\n
      * Each physic object has a group id which is attributed at creation.
      * The engine will only test collisions between somme registered groups. So if you want to receive event when object of group 1 intersect object from group 3, you have to call AddPhysicGroup(1, 3) or AddPhysicGroup(3, 1).\\n
      * Note: AddPhysicGroup(2, 2) is legal.
      */
    void AddPhysicGroup(int phGroupId1, int phGroupId2);

    // Background color
    /** Access the screen background color (also used for strip colors) */
    //LUA// luaProperty(Color, backgroundColor, GetBackgroundColor, SetBackgroundColor);

    /** Get the screen background color (also used for strip colors) */
    Color GetBackgroundColor();

    /** Set the screen background color (also used for strip colors) */
    void SetBackgroundColor(const Color &color);

    // Config
    /** Access the screen configuration */
    //LUA// luaProperty(ScreenConfig, screenConfig, GetScreenConfig, SetScreenConfig);

    /** Get the screen configuration */
    ScreenConfig &GetScreenConfig();

    /** Set the screen configuration */
    void SetScreenConfig(const ScreenConfig &screenConfig);

    /// LUA_EXPORT_OFF ///

    // No Lua
    // Lua
    //static void RegisterClassLua(lua_State *myLuaState);

    // Milcaneous
    void OnGetFocus();
    void OnLoseFocus();
    void ProceedEvent();

    void Run();
    virtual void Render(sf::RenderTarget &App);

    // Type
    ADD_GET_TYPE_FN(OBJ_TYPE_DRAWABLE_SCREEN)

    PhysicWorld *GetPhysicWorld() { return &m_world;}

protected:
    // Clone & Create
    /** Create a clone of the object and return an handler to the clone. */
    virtual Screen *Clone() const {return CloneToGroup(NULL);}

    /** Create a clone of the object, change the parent and return an handler to the clone. */
    virtual Screen *CloneToGroup(Group *newParent) const {Screen *obj = new Screen(*this); obj->RegisterClone(newParent); return obj;}
    void SetMouseRelative(bool relative, bool changeValue);

    void DrawRatioStrip(sf::RenderTarget &App);
    void RefreshStripView(sf::RenderTarget &App);
protected:
    PhysicWorld m_world;
    std::list<sf::Sound> m_sndTbl; // Il faut une liste pour ne pas détruire les sons en cas de resize !

    std::vector<EventHandler*> m_evtHandlerTbl;
    std::vector<PseudoEventHandler*> m_pseudoEvtHandlerTbl;

    // Mouse
    bool m_isMouseRelative;
    sf::Vector2<int> m_backMousePos;
    bool m_isCursorVisible;

    // FPS
    // bool m_showFps;

    // Time Speed
    float m_timeSpeed;

    // Background color
    Color m_backgroundColor;

    // View & strip & cfg
    sf::View m_screenView;
    sf::Shape m_strip1, m_strip2;

    ScreenConfig m_screenCfg;
    sf::Vector2i m_savedWindowSize;

private:
    virtual void ChangeParent(Group *parent); // Impossible de changer de parent
};
//------------------------------------------------------------------------------
} // namespace En

#endif // MAINGROUP_HPP
